<?php

namespace Drupal\graphql_fragment_include\GraphQL\Fragment;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;

/**
 * An implementation of GraphqlFragmentLoaderInterface.
 *
 * Allows inclusion of GraphQL Fragments on a query, defined by the
 * following syntax:
 * # include Fragment.gql
 * Includes are not supported by GraphQL, so we need to use comments for them.
 */
class GraphQLFragmentLoader implements GraphQLFragmentLoaderInterface {

  /**
   * Logger instance for this module.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * Config object for this module.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * Drupal file system.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * Loaded includes.
   *
   * @var array
   */
  protected $loadedFragments;

  /**
   * Constructor for IncludeLoader.
   *
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $loggerFactory
   *   Logger Factory.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   Drupal config service.
   * @param \Drupal\Core\File\FileSystemInterface $fileSystem
   *   Drupal file system.
   */
  public function __construct(
    LoggerChannelFactoryInterface $loggerFactory,
    ConfigFactoryInterface $configFactory,
    FileSystemInterface $fileSystem
  ) {
    $this->logger = $loggerFactory->get('graphql_fragment_include');
    $this->config = $configFactory->get('graphql_fragment_include.settings');
    $this->fileSystem = $fileSystem;
  }

  /**
   * {@inheritdoc}
   */
  public function findFragments(string $graphQLQuery) {
    preg_match_all("/^#\s+include\s+(\S+)\s*$/m", $graphQLQuery, $matches);
    return $matches;
  }

  /**
   * {@inheritdoc}
   */
  public function loadFragments(string $graphQLQuery) {
    $matches = $this->findFragments($graphQLQuery);
    if ($matches) {
      $baseDir = DRUPAL_ROOT . $this->config->get("fragments_base_dir");
      foreach ($matches[1] as $key => $match) {
        $fragmentPath = $this->fileSystem->realpath($baseDir . '/' . $match);
        if (!$this->isFragmentPathValid($fragmentPath)) {
          $this->logger->warning(t("Fragment path '%fragment-path' is not valid.", ['fragment-file' => $fragmentPath]));
          continue;
        }
        if (!is_file($fragmentPath)) {
          $this->logger->warning(t("No fragment '%fragment-path' found.", ['fragment-file' => $fragmentPath]));
          continue;
        }

        // Avoid double loading a fragment.
        if ($this->isFragmentAlreadyLoaded($fragmentPath)) {
          $graphQLQuery = trim(str_replace($matches[0][$key], "", $graphQLQuery));
        }
        else {
          $this->markFragmentAsLoaded($fragmentPath);
          $fragmentData = trim(file_get_contents($fragmentPath));
          // Recursively load other fragments.
          $fragmentData = $this->loadFragments($fragmentData);
          $graphQLQuery = trim(str_replace($matches[0][$key], "", $graphQLQuery));
          $graphQLQuery .= "\n\n# START $match \n" . $fragmentData . "\n# END $match ";
        }
      }
    }
    return $graphQLQuery;
  }

  /**
   * Tells whether a fragment path is valid.
   *
   * It's considered valid if it starts with fragments_base_dir.
   *
   * @param string $fragmentPath
   *   The file to check.
   *
   * @return bool
   *   True if it's valid, false otherwise.
   */
  protected function isFragmentPathValid(string $fragmentPath) {
    $baseDir = DRUPAL_ROOT . $this->config->get("fragments_base_dir");
    $fragmentPathStart = mb_substr($fragmentPath, 0, strlen($baseDir));
    if ($fragmentPathStart === $baseDir) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Tells whether a fragment has been already loaded.
   *
   * GraphQL fragments can not be repeated.
   *
   * @param string $fragmentPath
   *   The file to check.
   *
   * @return bool
   *   True if it has been already loaded, false otherwise.
   */
  protected function isFragmentAlreadyLoaded(string $fragmentPath) {
    return isset($this->loadedFragments[$fragmentPath]);
  }

  /**
   * Marks a fragment as already loaded.
   *
   * @param string $fragmentPath
   *   The file to be marked as already loaded.
   */
  protected function markFragmentAsLoaded(string $fragmentPath) {
    $this->loadedFragments[$fragmentPath] = TRUE;
  }

}
